import os
import sys
import cv2
import json
import uuid
import base64
import pydicom
import numpy as np
from io import BytesIO
from pydicom.dataset import Dataset
from PIL import ImageFont, Image, ImageDraw
import medical_image
from medical_image.utils import *


def init_json():
    return {
            "Height": None,
            "Width": None,
            "GraphicAnnotationSequence": [
                # {
                #     "GraphicLayer": None,
                #     "TextObjectSequence": [],
                #     "GraphicObjectSequence": [],
                #     "Probability": None,
                #     "Heatmap": None
                # }
            ],
            "SCArray": None,
            "ReportArray": None,
            "ImageComments": None,
            "Skipping": False
    }


def init_json_3D():
    return {
            "GraphicAnnotationSequence": [
                # {
                #     "GraphicLayer": None,
                #     "Objects": [],
                #     "Probability": None,
                # }
            ],
            "ReportArray": None,
            "ImageComments": None
    }


class DicomModule(object):
    def __init__(self, web_json):
        self.read_config(web_json)
        self.device = self.web_operation['device']
        self.gpu_index = self.web_operation['gpuIndex']
        self.init_models()

        self.app_name = os.environ["APP_NAME"]
        self.app_version = os.environ["APP_VERSION"]
        self.write_version = True

        self.graphic_layer_sequence = [
            get_graphic_layer_dataset('Logo', 0),
            get_graphic_layer_dataset('No Finding', 1)
        ]

        self.logo = Image.open(os.path.join(medical_image.__path__, 'logo/deepnoid.png'))
        self.colorbar = Image.open(os.path.join(medical_image.__path__, 'logo/colorbar.png'))

    def read_config(self, web_json):
        self.web_json = web_json
        self.web_vis = web_json['models']['visualization']
        self.web_sc = web_json['models']['secondary_capture']
        if 'report' in web_json['models']:
            self.web_report = web_json['models']['report']
        else:
            self.web_report = None
        if 'grayscale_softcopy_presentation_state' in web_json['models']:
            self.web_gsps = web_json['models']['grayscale_softcopy_presentation_state']
        else:
            self.web_gsps = None
        self.web_operation = web_json['configurations']['operation']
            
    def init_models(self):
        raise NotImplementedError

    def run(self, web_json, input_path, output_path):
        self.read_config(web_json)

        scanned_studies = scan_directory(input_path)
        study_uids = [key for key in scanned_studies.keys()]

        print(study_uids, flush=True)
        if len(study_uids) > 1:
            print(len(study_uids), flush=True)
            raise AssertionError
        study_uid = study_uids[0]

        study_paths = scanned_studies[study_uid]
        series_uids = [key for key in study_paths.keys()]

        ds_init = pydicom.dcmread(study_paths[series_uids[0]][0], force=True)
        
        if os.environ['AI_REQUEST_TAG'] == 'Private3355':
            if (0x3355, 0x1002) in ds_init:
                req_types = ds_init[0x3355, 0x1002].value.split('::')[-1].split('&')
                preferred_charsets = ds_init[0x3355, 0x1003].value.split('::')[-1]
            else:
                req_types = []
                if self.web_sc['@create']['selected'] == 'On':
                    req_types.append('Secondary-Capture')
                if self.web_gsps['@create']['selected'] == 'On':
                    req_types.append('GSPS')
                preferred_charsets = ''
        elif os.environ['AI_REQUEST_TAG'] == 'StudyComments':
            _, req_types, preferred_charsets = ds_init.StudyComments.split('::')[2::2]
            req_types = req_types.split('&')
        else:
            raise KeyError
        self.preferred_charsets = preferred_charsets.split('\\')[-1]
        print("req_types:", req_types)
        print("preferred_charsets:", self.preferred_charsets)

        referenced_series_sequence = []
        displayed_area_selection_sequence = []
        graphic_annotation_sequence = []
        VOI_LUT_sequence = []
        finding_sequences = []
        conversion_source_attributes_sequence = []

        self.generated_uuid = uuid.uuid4().int
        for series_uid in series_uids:
            series_paths = study_paths[series_uid]
            referenced_image_sequence = []
            finding_sequence = []
            for path in series_paths:
                ds = pydicom.dcmread(str(path), force=True)
                result_json = self.run_models(ds)
                if not result_json['Skipping']:
                    referenced_image = get_referenced_image_dataset(ds)
                    referenced_image_sequence.append(referenced_image)
                    conversion_source_attributes_sequence.append(referenced_image)
                    displayed_area_selection_sequence.append(get_displayed_area_selection(ds, [referenced_image]))
                    VOI_LUT_sequence.append(get_VOI_LUT_dataset(ds, [referenced_image]))

                    finding_sequence += self.update_gsps(graphic_annotation_sequence, referenced_image, result_json)
                    self.add_logo_annotation(graphic_annotation_sequence, referenced_image, result_json)

                    if 'Secondary-Capture' in req_types:
                        self.make_sc_dcm(ds, result_json, output_path)

            if len(finding_sequence):
                finding_sequences += finding_sequence
                referenced_series_sequence.append(get_referenced_series_dataset(series_uid, referenced_image_sequence))

        if len(finding_sequences) and ('GSPS' in req_types):
            self.make_gsps_dcm(
                ds,
                finding_sequences,
                displayed_area_selection_sequence,
                graphic_annotation_sequence,
                referenced_series_sequence,
                conversion_source_attributes_sequence,
                VOI_LUT_sequence,
                output_path
            )

    def make_sc_dcm(self, ds, result_json, output_path):
        # assert (type(result_json['InputArray']) == np.ndarray) or type(result_json['SCArray']) == np.ndarray
        referenced_image = get_referenced_image_dataset(ds)

        ds_sc = make_dataset(ds, self.app_name, self.app_version, 'SC', self.generated_uuid, self.preferred_charsets)
        ds_sc.WindowCenter = 128
        ds_sc.WindowWidth = 255
        if '@append_series_number' in self.web_sc:
            if self.web_sc['@append_series_number']['selected'] == 'On':
                ds_sc.SeriesNumber = self.web_sc['series_number'] + str(ds.SeriesNumber)
            else:
                ds_sc.SeriesNumber = self.web_sc['series_number']
        else:
            ds_sc.SeriesNumber = self.web_sc['series_number']

        if type(result_json["SCArray"]) == np.ndarray:
            ds_sc.PixelData = result_json["SCArray"].tobytes()
        else:
            assert "InputArray" not in result_json
            assert result_json["SCArray"] == None
            assert result_json['Height'] == ds.Rows
            assert result_json['Width'] == ds.Columns
            result_json['InputArray'] = read_dcm(ds)
            ds_sc.PixelData = self.make_sc_array(result_json).tobytes()
            
        ds_sc.Rows = result_json["Height"]
        ds_sc.Columns = result_json["Width"]

        if len(result_json["GraphicAnnotationSequence"]):
            finding_sequence = []
            for graphic_annotation in result_json["GraphicAnnotationSequence"]:
                dataset = Dataset()
                dataset.add_new((0x2021, 0x0010), 'LO', 'DEEPNOID')
                dataset.add_new((0x2021, 0x1009), 'UT', json.dumps({
                    'Name': graphic_annotation['GraphicLayer'],
                    'Probability': float(graphic_annotation['Probability'])
                }))
                dataset.ReferencedImageSequence = [referenced_image]
                finding_sequence.append(dataset)
            finding_sequence = sorted(finding_sequence, key=lambda seq: json.loads(seq[0x2021, 0x1009].value)['Probability'], reverse=True)
        else:
            dataset = Dataset()
            dataset.add_new((0x2021, 0x0010), 'LO', 'DEEPNOID')
            dataset.add_new((0x2021, 0x1009), 'UT', json.dumps({
                'Name': 'No Finding',
                'Probability': 0.0000
            }))
            dataset.ReferencedImageSequence = [referenced_image]
            finding_sequence = [dataset]

        set_private_tags(ds_sc, finding_sequence, self.app_name, self.app_version)
        if type(result_json['ImageComments']) == str:
            ds_sc.ImageComments = result_json['ImageComments']
        ds_sc.save_as(str(output_path / (ds_sc.SOPInstanceUID + '.dcm')))
        print(str(output_path / (ds_sc.SOPInstanceUID + '.dcm')))

    def make_sc_array(self, result_json):
        mfont = ImageFont.truetype(os.path.join(medical_image.__path__, 'fonts/NotoSansCJKkr-Medium.otf'), int(result_json['Height'] / 50))
        image = Image.fromarray(result_json['InputArray'].astype(np.uint8))
        img_editable = ImageDraw.Draw(image)

        ## Logo
        logo_w = self.logo.size[0] * result_json['Height'] * 0.55 / 2472
        logo_h = self.logo.size[1] * result_json['Height'] * 0.55 / 2472
        x = (result_json['Width'] - logo_w) / 2
        y = result_json['Height'] * 0.945 - logo_h
        paste_image(image, self.logo, x, y, logo_w, logo_h, 1)

        ## App Info
        msg = APP2PROUDCT[self.app_name]
        if self.write_version:
            msg += ' v.' + self.app_version
        mw, mh = mfont.getsize(msg)
        x = (result_json['Width'] - mw) / 2
        y = result_json['Height'] * 0.940 - logo_h - mh
        shadow = int(mh / 20)
        img_editable.text((x + shadow, y + shadow), msg, (167, 167, 167), mfont)
        img_editable.text((x, y), msg, (255, 255, 255), mfont)

        ## Colorbar
        heatmaps = [g['Heatmap'] for g in result_json["GraphicAnnotationSequence"]]
        # if len(heatmaps) != heatmaps.count(""):
        if len(heatmaps) != heatmaps.count(None):
            w, h = self.colorbar.size
            w *= result_json['Height'] * 0.4 / 2472 * 1000 / 1757
            h *= result_json['Height'] * 0.4 / 2472 * 1000 / 1757
            x = result_json['Width'] * 0.05
            y = result_json['Height'] * 0.945 - h
            paste_image(image, self.colorbar, x, y, w, h, 1)
            sfont = ImageFont.truetype(os.path.join(medical_image.__path__, 'fonts/NotoSansCJKkr-Medium.otf'), int(result_json['Height'] / 90))
            hw, _ = sfont.getsize('High')
            img_editable.text((x, y + h), 'Low', (255, 255, 255), sfont)
            img_editable.text((x + w - hw, y + h), 'High', (255, 255, 255), sfont)

        ## GraphicObject
        contour_image_array = np.array(image)
        for graphic_annotation in result_json["GraphicAnnotationSequence"]:
            heatmap_str = graphic_annotation['Heatmap']
            if heatmap_str != None:
                heatmap = np.array(Image.open(BytesIO(base64.b64decode(heatmap_str))))
                heatmap = cv2.cvtColor(heatmap, cv2.COLOR_BGRA2BGR)
                contour_image_array = cv2.addWeighted(contour_image_array, 1, heatmap, float(self.web_vis['opacity']), 0)

            for graphic_object in graphic_annotation['GraphicObjectSequence']:
                graphic_data = np.array(graphic_object['graphic_data'], np.int32)
                line_color = graphic_object['color'] if 'color' in graphic_object else (255, 255, 255)
                if ('shadow_color' not in graphic_object) or (graphic_object['shadow_color'] != 'OFF'):
                    shadow_color = graphic_object['shadow_color'] if 'shadow_color' in graphic_object else (0, 0, 0)
                    contour_image_array = cv2.polylines(contour_image_array, [graphic_data], False, shadow_color, 10)
                contour_image_array = cv2.polylines(contour_image_array, [graphic_data], False, line_color, max(int(min(result_json['Width'], result_json['Height']) / 500), 1))

        image = Image.fromarray(contour_image_array)
        img_editable = ImageDraw.Draw(image)
        ## TextObject
        for graphic_annotation in result_json["GraphicAnnotationSequence"]:
            for text_object in graphic_annotation['TextObjectSequence']:
                text_color = text_object['color'] if 'color' in text_object else (255, 255, 255)
                msg = text_object['text_data']
                x, y, _, _ = text_object['bbox']
                if ('shadow_color' not in text_object) or (text_object['shadow_color'] != 'OFF'):
                    mw, mh = mfont.getsize(msg)
                    shadow = int(mh / 20)
                    shadow_color = text_object['shadow_color'] if 'shadow_color' in text_object else (167, 167, 167)
                    img_editable.text((x + shadow, y + shadow), msg, shadow_color, mfont)
                img_editable.text((x, y), msg, text_color, mfont)
        return np.array(image)

    def update_gsps(self, graphic_annotation_sequence, referenced_image, result_json):
        def get_text_object_dcm(text_objects):
            text_object_sequence = []
            for text_object in text_objects:
                text_color = text_object['color'] if 'color' in text_object else [255, 255, 255]
                if 'shadow_color' not in text_object:
                    text_style = get_text_style_dataset(
                        color=text_color,
                        shadow_style='NORMAL',
                        shadow_color=[167, 167, 167],
                        shadow_offset=[int((text_object['bbox'][3] - text_object['bbox'][1]) / 20)] * 2
                    )
                elif text_object['shadow_color'] == 'OFF':
                    text_style = get_text_style_dataset(
                        color=text_color,
                        shadow_style='OFF'
                    )
                else:
                    text_style = get_text_style_dataset(
                        color=text_color,
                        shadow_style='NORMAL',
                        shadow_color=text_object['shadow_color'],
                        shadow_offset=[int((text_object['bbox'][3] - text_object['bbox'][1]) / 20)] * 2
                    )
                text_object_sequence.append(
                    get_text_object_dataset(
                        text_data=text_object["text_data"],
                        text_style=text_style,
                        bbox=text_object["bbox"]
                    )
                )
            return text_object_sequence

        def get_graphic_object_dcm(graphic_objects):
            graphic_object_sequence = []
            for graphic_object in graphic_objects:
                line_color = graphic_object['color'] if 'color' in graphic_object else [255, 255, 255]
                if 'shadow_color' not in graphic_object:
                    line_style = get_line_style_dataset(
                        color=line_color,
                        thickness=3.0,
                        shadow_style='OUTLINED',
                        shadow_color=[0, 0, 0],
                        shadow_offset=[6, 6]
                    )
                elif graphic_object['shadow_color'] == 'OFF':
                    line_style = get_line_style_dataset(
                        color=line_color,
                        thickness=3.0,
                        shadow_style='OFF'
                    )
                else:
                    line_style = get_line_style_dataset(
                        color=line_color,
                        thickness=3.0,
                        shadow_style='OUTLINED',
                        shadow_color=graphic_object['shadow_color'],
                        shadow_offset=[6, 6]
                    )
                graphic_object_sequence.append(
                    get_graphic_object_dataset(
                        graphic_data=list(graphic_object["graphic_data"].ravel()),
                        line_style=line_style
                    )
                )
            return graphic_object_sequence

        if len(result_json["GraphicAnnotationSequence"]):
            finding_sequence = []
            for graphic_annotation in result_json["GraphicAnnotationSequence"]:
                graphic_annotation_sequence.append(
                    get_graphic_annotation_dataset(
                        referenced_image_sequence=[referenced_image],
                        layer_name=graphic_annotation['GraphicLayer'],
                        text_object_sequence=get_text_object_dcm(graphic_annotation["TextObjectSequence"]),
                        graphic_object_sequence=get_graphic_object_dcm(graphic_annotation["GraphicObjectSequence"])
                    )
                )
                dataset = Dataset()
                dataset.add_new((0x2021, 0x0010), 'LO', 'DEEPNOID')
                if graphic_annotation['Heatmap'] == None:
                    heatmap_str = encoding_heatmap(np.zeros([result_json['Height'], result_json['Width']]), 0.5)
                    opacity = 0.5
                else:
                    heatmap_str = graphic_annotation['Heatmap']
                    opacity = self.web_vis['opacity']
                dataset.add_new((0x2021, 0x1009), 'UT', json.dumps({
                    'Name': graphic_annotation['GraphicLayer'],
                    'Probability': float(graphic_annotation['Probability']),
                    'Heatmap': heatmap_str,
                    'BlendingRatio': float(opacity)
                }))
                dataset.ReferencedImageSequence = [referenced_image]
                finding_sequence.append(dataset)

        else:
            dataset = Dataset()
            dataset.add_new((0x2021, 0x0010), 'LO', 'DEEPNOID')
            dataset.add_new((0x2021, 0x1009), 'UT', json.dumps({
                'Name': 'No Finding',
                'Probability': 0.0000,
                'Heatmap': encoding_heatmap(np.zeros([result_json['Height'], result_json['Width']]), 0.5),
                'BlendingRatio': float(0.5)
            }))
            dataset.ReferencedImageSequence = [referenced_image]
            finding_sequence = [dataset]
        return finding_sequence

    def add_logo_object(self, text_object_sequence, result_json):
        mfont = ImageFont.truetype(os.path.join(medical_image.__path__, 'fonts/NotoSansCJKkr-Medium.otf'), int(result_json['Height'] / 50))
        ## Logo
        msg = 'D E E P N O I D'
        w0, h0 = mfont.getsize(msg)
        x = (result_json['Width'] - w0) / 2
        y = result_json['Height'] * 0.945 - h0
        shadow = int(h0 / 20)
        text_style = get_text_style_dataset(
            color=[255, 255, 255],
            shadow_style='NORMAL',
            shadow_color=[255, 255, 255],
            shadow_offset=[shadow, shadow],
        )
        text_object_sequence.append(
            get_text_object_dataset(
                text_data=msg,
                text_style=text_style,
                bbox=[x, y, x + w0, y + h0]
            )
        )

        ## App
        msg = APP2PROUDCT[self.app_name]
        if self.write_version:
            msg += ' v.' + self.app_version
        w1, h1 = mfont.getsize(msg)
        x = (result_json['Width'] - w1) / 2
        y = result_json['Height'] * 0.940 - h0 - h1
        shadow = int(h1 / 20)
        text_style = get_text_style_dataset(
            color=[255, 255, 255],
            shadow_style='NORMAL',
            shadow_color=[255, 255, 255],
            shadow_offset=[shadow, shadow],
        )
        text_object_sequence.append(
            get_text_object_dataset(
                text_data=msg,
                text_style=text_style,
                bbox=[x, y, x + w1, y + h1]
            )
        )

    def add_logo_annotation(self, graphic_annotation_sequence, referenced_image, result_json):
        text_object_sequence = []
        self.add_logo_object(text_object_sequence, result_json)
        graphic_annotation = get_graphic_annotation_dataset(
            referenced_image_sequence=[referenced_image],
            layer_name='Logo',
            text_object_sequence=text_object_sequence,
            graphic_object_sequence=[]
        )
        graphic_annotation_sequence.append(graphic_annotation)

    def make_gsps_dcm(
        self,
        ds,
        finding_sequences,
        displayed_area_selection_sequence,
        graphic_annotation_sequence,
        referenced_series_sequence,
        conversion_source_attributes_sequence,
        VOI_LUT_sequence,
        output_path
        ):
        ds_gsps = make_dataset(ds, self.app_name, self.app_version, 'GSPS', self.generated_uuid, self.preferred_charsets)
        if '@append_series_number' in self.web_gsps:
            if self.web_gsps['@append_series_number']['selected'] == 'On':
                ds_gsps.SeriesNumber = self.web_gsps['series_number'] + str(ds.SeriesNumber)
            else:
                ds_gsps.SeriesNumber = self.web_gsps['series_number']
        else:
            ds_gsps.SeriesNumber = os.environ['GSPS_SERIES_NUMBER'] #self.web_gsps['series_number']
        ds_gsps.GraphicLayerSequence = self.graphic_layer_sequence

        if len(finding_sequences):
            finding_sequences = sorted(finding_sequences, key=lambda seq: json.loads(seq[0x2021, 0x1009].value)['Probability'], reverse=True)

        if ds.PhotometricInterpretation == 'MONOCHROME1':
            ds_gsps.PresentationLUTShape = 'INVERSE'

        set_private_tags(ds_gsps, finding_sequences, self.app_name, self.app_version)
        ds_gsps.DisplayedAreaSelectionSequence = displayed_area_selection_sequence
        ds_gsps.GraphicAnnotationSequence = graphic_annotation_sequence
        ds_gsps.ReferencedSeriesSequence = referenced_series_sequence
        ds_gsps.ConversionSourceAttributesSequence = conversion_source_attributes_sequence
        ds_gsps.SoftcopyVOILUTSequence = VOI_LUT_sequence

        ds_gsps.save_as(str(output_path / 'gsps.dcm'))
        print(str(output_path / 'gsps.dcm'))


class DicomModule3D(object):
    def __init__(self, web_json):
        self.read_config(web_json)
        self.device = self.web_operation['device']
        self.gpu_index = self.web_operation['gpuIndex']
        self.init_models()

        self.app_name = os.environ["APP_NAME"]
        self.app_version = os.environ["APP_VERSION"]
        self.write_version = True

        self.graphic_layer_sequence = [
            get_graphic_layer_dataset('Logo', 0),
            get_graphic_layer_dataset('No Finding', 1)
        ]

        self.logo = Image.open(os.path.join(medical_image.__path__, 'logo/deepnoid.png'))
        self.colorbar = Image.open(os.path.join(medical_image.__path__, 'logo/colorbar.png'))

    def read_config(self, web_json):
        self.web_json = web_json
        self.web_vis = web_json['models']['visualization']
        self.web_sc = web_json['models']['secondary_capture']
        if 'report' in web_json['models']:
            self.web_report = web_json['models']['report']
        else:
            self.web_report = None
        if 'grayscale_softcopy_presentation_state' in web_json['models']:
            self.web_gsps = web_json['models']['grayscale_softcopy_presentation_state']
        else:
            self.web_gsps = None
        # TODO 나중에 3D에서도 GSPS 적용할 수 있도록 수정
        # self.web_gsps = web_json['models']['grayscale_softcopy_presentation_state']
        self.web_operation = web_json['configurations']['operation']
            
    def init_models(self):
        raise NotImplementedError

    def run(self, web_json, input_path, output_path):
        self.read_config(web_json)
        if (self.device != self.web_operation['device']) or (self.gpu_index != self.web_operation['gpuIndex']):
            print('Configuration of Device has been changed. Reload model.', flush=True)
            self.__delattr__('models')
            self.device = self.web_operation['device']
            self.gpu_index = self.web_operation['gpuIndex']
            self.init_models()

        scanned_studies = scan_directory(input_path)
        study_uids = [key for key in scanned_studies.keys()]

        print(study_uids, flush=True)
        if len(study_uids) > 1:
            print(len(study_uids), flush=True)
            raise AssertionError
        study_uid = study_uids[0]

        study_paths = scanned_studies[study_uid]
        series_uids = [key for key in study_paths.keys()]

        ds_init = pydicom.dcmread(study_paths[series_uids[0]][0], force=True)
        
        if os.environ['AI_REQUEST_TAG'] == 'Private3355':
            if (0x3355, 0x1002) in ds_init:
                req_types = ds_init[0x3355, 0x1002].value.split('::')[-1].split('&')
                preferred_charsets = ds_init[0x3355, 0x1003].value.split('::')[-1]
            else:
                req_types = []
                if self.web_sc['@create']['selected'] == 'On':
                    req_types.append('Secondary-Capture')
                if self.web_report and self.web_report['@create']['selected'] == 'On':
                    req_types.append('Report')
                # if self.web_gsps['@create']['selected'] == 'On':
                #     req_types.append('GSPS')
                preferred_charsets = ''
        elif os.environ['AI_REQUEST_TAG'] == 'StudyComments':
            _, req_types, preferred_charsets = ds_init.StudyComments.split('::')[2::2]
            req_types = req_types.split('&')
            if 'GSPS' in req_types:
                # TODO 나중에 3D에서도 GSPS 적용할 수 있도록 수정
                raise NotImplementedError
        else:
            raise KeyError
        self.preferred_charsets = preferred_charsets.split('\\')[-1]
        print("req_types:", req_types)
        print("preferred_charsets:", self.preferred_charsets)

        referenced_series_sequence = []
        displayed_area_selection_sequence = []
        graphic_annotation_sequence = []
        VOI_LUT_sequence = []
        finding_sequences = []
        conversion_source_attributes_sequence = []

        self.generated_uuid = uuid.uuid4().int
        for series_uid in series_uids:
            series_paths = study_paths[series_uid]
            referenced_image_sequence = []
            finding_sequence = []
            ds_dict = {'path': series_paths[0].parent}
            for path in series_paths:
                ds = pydicom.dcmread(str(path), force=True)
                ds_dict[int(ds.InstanceNumber)] = ds

                referenced_image = get_referenced_image_dataset(ds)
                referenced_image_sequence.append(referenced_image)
                conversion_source_attributes_sequence.append(referenced_image)
                displayed_area_selection_sequence.append(get_displayed_area_selection(ds, [referenced_image]))
                VOI_LUT_sequence.append(get_VOI_LUT_dataset(ds, [referenced_image]))

                # finding_sequence = self.update_gsps(graphic_annotation_sequence, referenced_image, result_json)
                # self.add_logo_annotation(graphic_annotation_sequence, referenced_image, result_json)


            # finding_sequences += finding_sequence
            # referenced_series_sequence.append(get_referenced_series_dataset(series_uid, referenced_image_sequence))

            result_json = self.run_models(ds_dict)
            
            if 'Secondary-Capture' in req_types:
                self.make_sc_dcm(ds_dict, result_json, output_path)

            if 'Report' in req_types:
                report_array = result_json['ReportArray']
                ds_rp = make_dataset(ds, self.app_name, self.app_version, 'RP', self.generated_uuid, self.preferred_charsets)
                ds_rp.PixelData = report_array.tobytes()
                ds_rp.Rows = report_array.shape[0]
                ds_rp.Columns = report_array.shape[1]
                ds_rp.save_as(str(output_path / (ds_rp.SOPInstanceUID + '.dcm')))
                # print(str(output_path / (ds_rp.SOPInstanceUID + '.dcm')))

        if 'GSPS' in req_types:
            # TODO 나중에 3D에서도 GSPS 적용할 수 있도록 수정
            raise NotImplementedError

    def make_sc_dcm(self, ds_dict, result_json, output_path):
        # assert (type(result_json['InputArray']) == np.ndarray) or type(result_json['SCArray']) == np.ndarray

        instance_numbers = {}
        for i, graphic_annotation in enumerate(result_json["GraphicAnnotationSequence"]):
            objects = graphic_annotation['Objects']
            _ins = [obj['instance_number'] for obj in objects]
            for _in in _ins:
                if _in in instance_numbers:
                    instance_numbers[_in][i] = _ins.index(_in)
                else:
                    instance_numbers[_in] = {i: _ins.index(_in)}

        for instance_number in [key for key in ds_dict.keys() if type(key) == int]:
            ds = ds_dict[instance_number]
            referenced_image = get_referenced_image_dataset(ds)

            ds_sc = make_dataset(ds, self.app_name, self.app_version, 'SC', self.generated_uuid, self.preferred_charsets)
            ds_sc.WindowCenter = 128
            ds_sc.WindowWidth = 255
            if '@append_series_number' in self.web_sc:
                if self.web_sc['@append_series_number']['selected'] == 'On':
                    ds_sc.SeriesNumber = self.web_sc['series_number'] + str(ds.SeriesNumber)
                else:
                    ds_sc.SeriesNumber = self.web_sc['series_number']
            else:
                ds_sc.SeriesNumber = self.web_sc['series_number']

            objects = []
            if instance_number in instance_numbers:
                for i in instance_numbers[instance_number]:
                    objects.append(result_json["GraphicAnnotationSequence"][i]['Objects'][instance_numbers[instance_number][i]])
            ds_sc.PixelData = self.make_sc_array(ds, objects).tobytes()

            if len(objects):
                ds_sc.Rows = objects[0]['Height']
                ds_sc.Columns = objects[0]['Width']
            else:
                ds_sc.Rows = int(ds.Rows)
                ds_sc.Columns = int(ds.Columns)

            if len(result_json["GraphicAnnotationSequence"]):
                finding_sequence = []
                for i, graphic_annotation in enumerate(result_json["GraphicAnnotationSequence"]):
                    dataset = Dataset()
                    dataset.add_new((0x2021, 0x0010), 'LO', 'DEEPNOID')
                    dataset.add_new((0x2021, 0x1009), 'UT', json.dumps({
                        'Name': graphic_annotation['GraphicLayer'],
                        'Probability': float(graphic_annotation['Probability'])
                    }))
                    dataset.ReferencedImageSequence = [referenced_image]
                    finding_sequence.append(dataset)
                finding_sequence = sorted(finding_sequence, key=lambda seq: json.loads(seq[0x2021, 0x1009].value)['Probability'], reverse=True)
            else:
                dataset = Dataset()
                dataset.add_new((0x2021, 0x0010), 'LO', 'DEEPNOID')
                dataset.add_new((0x2021, 0x1009), 'UT', json.dumps({
                    'Name': 'No Finding',
                    'Probability': 0.0000
                }))
                dataset.ReferencedImageSequence = [referenced_image]
                finding_sequence = [dataset]

            set_private_tags(ds_sc, finding_sequence, self.app_name, self.app_version)
            if type(result_json['ImageComments']) == str:
                ds_sc.ImageComments = result_json['ImageComments']
            ds_sc.save_as(str(output_path / (ds_sc.SOPInstanceUID + '.dcm')))
            # print(str(output_path / (ds_sc.SOPInstanceUID + '.dcm')))

    def update_gsps(self, *args, **kwargs):
        raise NotImplementedError

    def make_sc_array(self, ds, objects):
        if len(objects):
            height = objects[0]['Height']
            width = objects[0]['Width']
        else:
            height = int(ds.Rows)
            width = int(ds.Columns)
        mfont = ImageFont.truetype(os.path.join(medical_image.__path__, 'fonts/NotoSansCJKkr-Medium.otf'), int(height / 50))
        # image = Image.fromarray(objects['InputArray'].astype(np.uint8))
        image = Image.fromarray(read_dcm(ds).astype(np.uint8))
        img_editable = ImageDraw.Draw(image)

        ## Logo
        logo_w = self.logo.size[0] * height * 0.55 / 2472
        logo_h = self.logo.size[1] * height * 0.55 / 2472
        x = (width - logo_w) / 2
        y = height * 0.945 - logo_h
        paste_image(image, self.logo, x, y, logo_w, logo_h, 1)

        ## App Info
        msg = APP2PROUDCT[self.app_name]
        if self.write_version:
            msg += ' v.' + self.app_version
        mw, mh = mfont.getsize(msg)
        x = (width - mw) / 2
        y = height * 0.940 - logo_h - mh
        shadow = int(mh / 20)
        img_editable.text((x + shadow, y + shadow), msg, (167, 167, 167), mfont)
        img_editable.text((x, y), msg, (255, 255, 255), mfont)

        for obj in objects:
            ## GraphicObject
            contour_image_array = np.array(image)
            for graphic_object in obj['GraphicObjectSequence']:
                graphic_data = np.array(graphic_object['graphic_data'], np.int32)
                line_color = graphic_object['color'] if 'color' in graphic_object else (255, 255, 255)
                if ('shadow_color' not in graphic_object) or (graphic_object['shadow_color'] != 'OFF'):
                    shadow_color = graphic_object['shadow_color'] if 'shadow_color' in graphic_object else (0, 0, 0)
                    contour_image_array = cv2.polylines(contour_image_array, [graphic_data], False, shadow_color, 10)
                contour_image_array = cv2.polylines(contour_image_array, [graphic_data], False, line_color, max(int(min(width, height) / 500), 1))

            image = Image.fromarray(contour_image_array)
            img_editable = ImageDraw.Draw(image)
            ## TextObject
            for text_object in obj['TextObjectSequence']:
                text_color = text_object['color'] if 'color' in text_object else (255, 255, 255)
                msg = text_object['text_data']
                x, y, _, _ = text_object['bbox']
                if ('shadow_color' not in text_object) or (text_object['shadow_color'] != 'OFF'):
                    mw, mh = mfont.getsize(msg)
                    shadow = int(mh / 20)
                    shadow_color = text_object['shadow_color'] if 'shadow_color' in text_object else (167, 167, 167)
                    img_editable.text((x + shadow, y + shadow), msg, shadow_color, mfont)
                img_editable.text((x, y), msg, text_color, mfont)

        return np.array(image)

    def add_logo_object(self, text_object_sequence, result_json):
        mfont = ImageFont.truetype(os.path.join(medical_image.__path__, 'fonts/NotoSansCJKkr-Medium.otf'), int(result_json['Height'] / 50))
        ## Logo
        msg = 'D E E P N O I D'
        w0, h0 = mfont.getsize(msg)
        x = (result_json['Width'] - w0) / 2
        y = result_json['Height'] * 0.945 - h0
        shadow = int(h0 / 20)
        text_style = get_text_style_dataset(
            color=[255, 255, 255],
            shadow_style='NORMAL',
            shadow_color=[255, 255, 255],
            shadow_offset=[shadow, shadow],
        )
        text_object_sequence.append(
            get_text_object_dataset(
                text_data=msg,
                text_style=text_style,
                bbox=[x, y, x + w0, y + h0]
            )
        )

        ## App
        msg = APP2PROUDCT[self.app_name]
        if self.write_version:
            msg += ' v.' + self.app_version
        w1, h1 = mfont.getsize(msg)
        x = (result_json['Width'] - w1) / 2
        y = result_json['Height'] * 0.940 - h0 - h1
        shadow = int(h1 / 20)
        text_style = get_text_style_dataset(
            color=[255, 255, 255],
            shadow_style='NORMAL',
            shadow_color=[255, 255, 255],
            shadow_offset=[shadow, shadow],
        )
        text_object_sequence.append(
            get_text_object_dataset(
                text_data=msg,
                text_style=text_style,
                bbox=[x, y, x + w1, y + h1]
            )
        )

    def add_logo_annotation(self, graphic_annotation_sequence, referenced_image, result_json):
        text_object_sequence = []
        self.add_logo_object(text_object_sequence, result_json)
        graphic_annotation = get_graphic_annotation_dataset(
            referenced_image_sequence=[referenced_image],
            layer_name='Logo',
            text_object_sequence=text_object_sequence,
            graphic_object_sequence=[]
        )
        graphic_annotation_sequence.append(graphic_annotation)

    def make_gsps_dcm(self, *args,**kwargs):
        raise NotImplementedError



import os
import cv2
import copy
import json
import uuid
import base64
import pydicom
import numpy as np
from datetime import datetime
from pydicom.dataset import Dataset
from pydicom.valuerep import DSfloat
from pydicom.sequence import Sequence
from pydicom.multival import MultiValue
from pydicom.pixel_data_handlers.util import convert_color_space, apply_color_lut
from pydicom.pixel_data_handlers.util import apply_voi_lut


APP2PROUDCT = {
    'dc-xr-03': 'DEEP:CHEST-XR-03',
    'dc-xr-04': 'DEEP:CHEST-XR-04',
    'ds-cf-01': 'DEEP:SPINE-CF-01',
    'ds-sc-01': 'DEEP:SPINE-SC-01',
    'ds-xr-01': 'DEEP:SPINE-XR-01',
    'dn-ca-01': 'DEEP:NEURO-CA-01',
    'dl-ln-01': 'DEEP:LUNG-LN-01',
    'db-fr-01': 'DEEP:BONE-FR-01',
}


def encoding_heatmap(mask, thres=0.5):
    assert mask.max() <= 1
    assert mask.min() >= 0

    grad_cam = mask
    heatmap = cv2.applyColorMap(np.uint8(255 * np.clip(grad_cam - thres, 0, 1) / (1 - thres)), cv2.COLORMAP_JET)
    back_index = np.where(grad_cam < thres)
    heatmap[back_index] = 0
    b, g, r = cv2.split(heatmap)
    a = np.ones(b.shape, b.dtype) * 255
    a[back_index] = 0
    img_bgra = cv2.merge([b, g, r, a])
    cv2.imwrite('tmp.png', img_bgra)
    heatmap_encoded = base64.b64encode(open("tmp.png", "rb").read()).decode()
    os.remove('tmp.png')
    return heatmap_encoded


def rgb2Lab(color):
    color_Lab = cv2.cvtColor(np.array([[color]]).astype(np.uint8), cv2.COLOR_RGB2Lab).squeeze()
    return [color_Lab[0] * 256, color_Lab[1] * 256, color_Lab[2] * 256]


def paste_image(report, cam, x, y, w, h, scale):
    x1 = int(x * scale)
    y1 = int(y * scale)
    w = int(w * scale)
    h = int(h * scale)
    x2 = x1 + w
    y2 = y1 + h
    paste_image = cam.resize((w, h))
    if cam.mode == 'RGBA':
        report.paste(paste_image, [x1, y1, x2, y2], paste_image)
    else:
        report.paste(paste_image, [x1, y1, x2, y2])

def read_dcm(ds):
    arr = ds.pixel_array
    mode = ds.PhotometricInterpretation
    if 'YBR_FULL' in mode:
        ori_image_array = convert_color_space(arr, mode, 'RGB')
    elif mode == 'PALETTE COLOR':
        ori_image_array = apply_color_lut(arr, ds)
    elif 'MONOCHROME' in mode:
        try:
            if ds.__contains__('RescaleIntercept') and ds.__contains__('RescaleSlope'):
                intercept = ds.RescaleIntercept
                slope = ds.RescaleSlope
            else:
                intercept = 0
                slope = 1
            out = slope * arr + intercept
        except:
            out = arr
            
        if ds.__contains__('WindowWidth') and ds.__contains__('WindowCenter'):
            if type(ds.WindowCenter) == DSfloat:
                wc = ds.WindowCenter
                ww = ds.WindowWidth
            else:
                wc = ds.WindowCenter[0]
                ww = ds.WindowWidth[0]

            wl = (wc - 0.5 - (ww - 1) / 2)
            wu = (wc - 0.5 + (ww - 1) / 2)

            out[np.where(out <= wl)] = wl
            out[np.where(out >= wu)] = wu
            
            ori_image_array = (out - wl) / (wu - wl)
            
        elif ds.__contains__('VOILUTSequence'):
            out = apply_voi_lut(out, ds)
            ori_image_array = out / ds.VOILUTSequence[0].LUTDescriptor[0]
            
        else:
            raise KeyError

        if mode == 'MONOCHROME1':
            ori_image_array = 1 - ori_image_array

        ori_image_array *= 255

    if ori_image_array.ndim != 3:
        ori_image_array = np.stack([ori_image_array] * 3, axis=-1)
    if ori_image_array.shape[-1] == 4:
        ori_image_array = ori_image_array[..., :3]

    return ori_image_array


def scan_directory(input_path): 
    paths = sorted(list(input_path.glob('*.dcm')))
    study_instance_uids = {}
    for path in paths:
        ds = pydicom.dcmread(str(path), force=True)
        study_uid = ds.StudyInstanceUID
        series_uid = ds.SeriesInstanceUID
        if study_uid in study_instance_uids:
            if series_uid in study_instance_uids[study_uid]:
                study_instance_uids[study_uid][series_uid].append(path)
            else:
                study_instance_uids[study_uid].update(
                    {
                        series_uid: [path]
                    }
                )

        else:
            study_instance_uids.update(
                {
                    study_uid: {
                        series_uid: [path]
                    }
                }
            )
    return study_instance_uids


def make_dataset(input_dicom, app_name, app_version, result_dcm_format='SC', use_uid=None, preferred_charset=''):
    ds = Dataset()
    copy_dicom_tags(ds, copy.deepcopy(input_dicom), result_dcm_format)
    set_dicom_tags(ds, app_name, app_version, result_dcm_format)
    if preferred_charset:
        convert_charset(ds, preferred_charset)
    set_time(ds, result_dcm_format)
    set_uids(ds, input_dicom, use_uid, result_dcm_format)
    ds.preamble = input_dicom.preamble
    ds.file_meta.MediaStorageSOPClassUID = ds.SOPClassUID
    ds.file_meta.MediaStorageSOPInstanceUID = ds.SOPInstanceUID
    return ds


def convert_charset(dicom, preferred_charset):
    python_encode_type = pydicom.charset.convert_encodings(dicom.SpecificCharacterSet)[0]
    python_decode_type = pydicom.charset.convert_encodings(preferred_charset)[0]
    for tag in ['InstitutionName', 'PatientName', 'StudyDescription']:
        if dicom.__contains__(tag):
            raw_value = dicom[tag].value
            encoded_value = raw_value.encode(python_encode_type)
            decoded_value = encoded_value.decode(python_decode_type)
            setattr(dicom, tag, decoded_value)


def copy_dicom_tags(dicom, input_dicom, result_dcm_format='SC'):
    type1_elements = [
        'StudyInstanceUID', # General Study
        # 'SpecificCharacterSet',
    ]

    type2_elements = [
        'PatientName', # Patient
        'PatientID',
        'PatientBirthDate',
        'PatientSex',
        'StudyDate', # General Study
        'StudyTime',
        'ReferringPhysicianName',
        'StudyID',
        'AccessionNumber',
        'StudyDescription', # for SC
    ]
    if result_dcm_format == 'SC':
        type2_elements += [
            'Rows',
            'Columns',
            'PixelSpacing',
        ]
    elif result_dcm_format == 'GSPS':
        type2_elements += [
        ]

    type3_elements = [
        'PatientAge', # Patient Study
        'PatientSize',
        'PatientWeight',
        'InstitutionName',
        'StudyComments',
    ]
    if result_dcm_format == 'SC':
        type3_elements += [
            'ViewPosition',
            'Laterality',
            'ImagerPixelSpacing',
        ]
    elif result_dcm_format == 'GSPS':
        type3_elements += [
        ]
    private_elements = [
        (0x3355, 0x0010), # Priave Creator
        (0x3355, 0x1001), # AET
        (0x3355, 0x1002), # DEEPAIReqType
        (0x3355, 0x1003), # DEEPAIReqCharset
    ]

    dicom.file_meta = input_dicom.file_meta
    for tag in type1_elements:
        if dicom.__contains__(tag):
            assert dicom[tag].value in ['', input_dicom[tag].value]
        else:
            assert input_dicom.__contains__(tag)
            dicom.add(input_dicom.data_element(tag))

    tag = 'SpecificCharacterSet'
    if input_dicom.__contains__(tag):
        if type(input_dicom.SpecificCharacterSet) == MultiValue:
            dicom.SpecificCharacterSet = input_dicom.SpecificCharacterSet[-1]
        else:
            dicom.add(input_dicom.data_element(tag))
    else:
        dicom.SpecificCharacterSet = 'ISO_IR 100'

    for tag in type2_elements:
        if dicom.__contains__(tag):
            assert dicom[tag].value in ['', input_dicom[tag].value]
        else:
            if input_dicom.__contains__(tag):
                dicom.add(input_dicom.data_element(tag))
            else:
                setattr(dicom, tag, '')

    for tag in type3_elements:
        if dicom.__contains__(tag):
            assert dicom[tag].value == input_dicom[tag].value
        else:
            if input_dicom.__contains__(tag):
                dicom.add(input_dicom.data_element(tag))

    for tag in private_elements:
        if tag in dicom:
            assert dicom[tag].value == input_dicom[tag].value
        else:
            if tag in input_dicom:
                dicom[tag] = input_dicom[tag]

def set_uids(
        dicom,
        input_dicom,
        use_uid=None,
        result_dcm_format='SC'
    ):
    if not use_uid:
        use_uid = uuid.uuid4().int
    try:
        series_idx = int(input_dicom.SeriesNumber)
    except:
        series_idx = 1
    try:
        instance_idx = int(input_dicom.InstanceNumber)
    except:
        instance_idx = 1

    if result_dcm_format == 'SC':
        dicom.SeriesInstanceUID = '2.25.{:d}.1.{:d}'.format(series_idx, use_uid)
        dicom.SOPInstanceUID = '2.25.{:d}.1.{:d}.{:d}'.format(series_idx, instance_idx, use_uid)
        dicom.SeriesNumber = '99991' + str(series_idx)
        dicom.InstanceNumber = str(instance_idx)
    if result_dcm_format == 'RP':
        dicom.SeriesInstanceUID = '2.25.{:d}.2.{:d}'.format(series_idx, use_uid)
        dicom.SOPInstanceUID = '2.25.{:d}.2.{:d}.{:d}'.format(series_idx, instance_idx, use_uid)
        dicom.SeriesNumber = '99992' + str(series_idx)
        dicom.InstanceNumber = str(instance_idx)
    elif result_dcm_format == 'GSPS':
        dicom.SeriesInstanceUID = '2.25.0.0.{:d}'.format(use_uid)
        dicom.SOPInstanceUID = '2.25.0.0.0.{:d}'.format(use_uid)
        dicom.SeriesNumber = '999999999'
        dicom.InstanceNumber = '1'


def set_dicom_tags(dicom, app_name, app_version, result_dcm_format='SC'):
    series_tag_values = [
        ('SeriesDescription', APP2PROUDCT[app_name]),
        ('Manufacturer', 'DEEPNOID'), # General Equipment
    ]
    if result_dcm_format in ['SC', 'RP']:
        series_tag_values += [
            ('Modality', 'OT'),  # General Series
            ('SOPClassUID', '1.2.840.10008.5.1.4.1.1.7'), # SOP Common,
            ('SamplesPerPixel', 3),
            ('PhotometricInterpretation', 'RGB'),
            ('BitsAllocated', 8),
            ('BitsStored', 8),
            ('HighBit', 7),
            ('PlanarConfiguration', 0),
            ('PixelRepresentation', 0), # 0: unsigned integer, 1: 2's complement
            ('RescaleSlope', '1'),
            ('RescaleIntercept', '0'),
            ('RescaleType', 'US'), # Unspecified
            ('ConversionType', 'WSD'), # Workstation, The equipment that placed the modified image into a DICOM SOP instance.
        ]
    elif result_dcm_format == 'GSPS':
        series_tag_values += [
            ('InstanceNumber', 1),
            ('Modality', 'PR'),
            ('SOPClassUID', "1.2.840.10008.5.1.4.1.1.11.1"), # SOP Common,
            ('SoftwareVersions', app_version),
            ('ContentLabel', app_name),
            ('ContentDescription', 'Detect Consolidation and Pneumothorax from CXR.'),
            ('ContentCreatorName', 'DEEPNOID'),
            ('PresentationLUTShape', 'IDENTITY'),
        ]

    for tag, value in series_tag_values:
        if dicom.__contains__(tag):
            assert dicom[tag].value == value
        else:
            setattr(dicom, tag, value)
              
    dicom.file_meta.TransferSyntaxUID = '1.2.840.10008.1.2.1'


def set_time(dicom, result_dcm_format='SC'):
    creation_date = datetime.now().date().strftime("%Y%m%d")
    creation_time = datetime.now().time().strftime("%H%M%S")
    series_tag_values = [
        ('SeriesDate', creation_date),
        ('SeriesTime', creation_time),
    ]
    if result_dcm_format in ['SC', 'RP']:
        series_tag_values += [
            ('DateOfSecondaryCapture', creation_date), # SC Image Module
            ('TimeOfSecondaryCapture', creation_time),
        ]
    elif result_dcm_format == 'GSPS':
        series_tag_values += [
            ('PresentationCreationDate', creation_date), # Presentation State Identification
            ('PresentationCreationTime', creation_time),
        ]
    for tag, value in series_tag_values:
        setattr(dicom, tag, value)


def set_private_tags(dicom, finding_sequence, app_name, app_version):
    result_js = {
        'FormatVersion': '1.0',
        'Date': dicom.SeriesDate,
        'Time': dicom.SeriesTime,
        'Vendor': 'DEEPNOID',
        'Service': 'DEEP:AI',
        'Model': app_name,
        'Version': app_version
    }
    block = dicom.private_block(0x2021, 'DEEPNOID', create=True)
    block.add_new(0x01, 'LO', "DEEPNOID")
    block.add_new(0x02, 'LO', 'DEEP:AI')
    block.add_new(0x03, 'LO', app_name)
    block.add_new(0x04, 'LO', app_version)
    block.add_new(0x05, 'SS', 1)
    block.add_new(0x06, 'SS', 1)
    block.add_new(0x08, 'UT', json.dumps(result_js))
    block.add_new(0x11, 'SQ', Sequence(finding_sequence))

def get_referenced_image_dataset(input_dicom):
    dataset = Dataset()
    dataset.ReferencedSOPClassUID = input_dicom.SOPClassUID # (0008:1150)
    dataset.ReferencedSOPInstanceUID = input_dicom.SOPInstanceUID # (0008:1155)
    return dataset


def get_referenced_series_dataset(series_instance_uid, referenced_image_sequence):
    dataset = Dataset()
    dataset.SeriesInstanceUID = series_instance_uid
    dataset.ReferencedImageSequence = referenced_image_sequence
    return dataset


def get_graphic_layer_dataset(layer_name, layer_order):
    assert type(layer_name) == str
    assert type(layer_order) == int
    dataset = Dataset()
    dataset.GraphicLayer = layer_name
    dataset.GraphicLayerOrder = layer_order
    return dataset


def get_text_style_dataset(
    style='SOLID',
    color=[255, 255, 255],
    horizontal_align='CENTER',
    vertical_align='CENTER',
    shadow_style='NORMAL',
    shadow_color=[127, 127, 127],
    shadow_offset=[3, 3],
    shadow_opacity=1.0,
    ):
    
    assert style in ['SOLID', 'DASHED']
    assert shadow_style in ['NORMAL', 'OUTLINED', 'OFF']
    
    dataset = Dataset()
    dataset.CSSFontName = 'Malgun Gothic'
    dataset.TextColorCIELabValue = rgb2Lab(color)
    dataset.HorizontalAlignment = horizontal_align
    dataset.VerticalAlignment = vertical_align
    dataset.Underlined = 'N'
    dataset.Bold = 'N'
    dataset.Italic = 'N'
    dataset.ShadowStyle = shadow_style
    if shadow_style != 'OFF':
        dataset.ShadowColorCIELabValue = rgb2Lab(shadow_color)
        dataset.ShadowOffsetX = shadow_offset[0]
        dataset.ShadowOffsetY = shadow_offset[1]
        dataset.ShadowOpacity = shadow_opacity
    return dataset

def get_line_style_dataset(
    color,
    thickness,
    style='SOLID',
    opacity=1.0,
    off_color=[0, 0, 0],
    off_opacity=0.0,
    shadow_style='OUTLINED',
    shadow_color=[127, 127, 127],
    shadow_offset=[3, 3],
    shadow_opacity=1.0
    ):
    
    assert style in ['SOLID', 'DASHED']
    assert shadow_style in ['NORMAL', 'OUTLINED', 'OFF']
    
    dataset = Dataset()
    dataset.PatternOnColorCIELabValue = rgb2Lab(color)
    dataset.PatternOnOpacity = opacity
    dataset.LineThickness = thickness
    dataset.LineDashingStyle = style
    if style == 'DASHED':
        dataset.PaternOffColorCIELabValue = rgb2Lab(off_color)
        dataset.PatternOffOpacity = off_opacity
        # dataset.LinePattern = 00FFH 
    dataset.ShadowStyle = shadow_style
    dataset.ShadowColorCIELabValue = rgb2Lab(shadow_color)
    dataset.ShadowOffsetX = shadow_offset[0]
    dataset.ShadowOffsetY = shadow_offset[1]
    dataset.ShadowOpacity = shadow_opacity
    return dataset

def get_fill_style_dataset(
        color,
        off_color=[127, 127, 127],
        opacity=1.0,
        off_opacity=0.0,
        fill_mode='SOLID'
    ):
    assert fill_mode in ['SOLID', 'STIPPELED']
    dataset = Dataset()
    dataset.PatternOnColorCIELabValue = rgb2Lab(color)
    dataset.PatternOnOpacity = opacity
    dataset.FillMode = fill_mode
    if fill_mode == 'STIPPELED':
        dataset.PatternOffColorCIELabValue = rgb2Lab(off_color)
        dataset.PatternOffOpacity = off_opacity
        # dataset.FillPattern = 
    return dataset


def get_graphic_annotation_dataset(
        referenced_image_sequence,
        layer_name,
        text_object_sequence=None,
        graphic_object_sequence=None
    ):
    assert (text_object_sequence is not None) or (graphic_object_sequence is not None)

    dataset = Dataset()
    dataset.ReferencedImageSequence = referenced_image_sequence
    dataset.GraphicLayer = layer_name
    if text_object_sequence is not None:
        dataset.TextObjectSequence = text_object_sequence
    if graphic_object_sequence is not None:
        dataset.GraphicObjectSequence = graphic_object_sequence
    return dataset

def get_text_object_dataset(
        text_data,
        bbox, # (x1, y1, x2, y2)
        text_style,
        boundingbox_annotation_units='PIXEL',
        horizontal_justification='CENTER'
    ):
    dataset = Dataset()
    dataset.BoundingBoxAnnotationUnits = boundingbox_annotation_units
    dataset.BoundingBoxTopLeftHandCorner = [bbox[0], bbox[1]]
    dataset.BoundingBoxBottomRightHandCorner = [bbox[2], bbox[3]]
    dataset.BoundingBoxTextHorizontalJustification = horizontal_justification

    dataset.TextStyleSequence = Sequence([text_style])
    dataset.UnformattedTextValue = text_data
    return dataset

def get_displayed_area_selection(ds, referenced_image_sequence):
    dataset = Dataset()
    dataset.ReferencedImageSequence = referenced_image_sequence
    dataset.PixelOriginInterpretation = 'FRAME'
    dataset.DisplayedAreaTopLeftHandCorner = [1, 1]
    dataset.DisplayedAreaBottomRightHandCorner = [ds.Columns, ds.Rows]
    dataset.PresentationSizeMode = 'SCALE TO FIT'
    dataset.PresentationPixelSpacing = ds.PixelSpacing if 'PixelSpacing' in ds else [1, 1]
    return dataset

def get_graphic_object_dataset(
        graphic_data,
        line_style,
        fill_style=None,
        graphic_annotation_units='PIXEL',
        graphic_dimensions=2,
        graphic_type='POLYLINE',
        graphic_filled='N'
    ):
    assert graphic_annotation_units in ['PIXEL', 'DISPLAY', 'MATRIX']
    assert graphic_type in ['POINT', 'POLYLINE', 'INTERPOLATED', 'CIRCLE', 'ELLIPSE']
    assert graphic_filled in ['Y', 'N']
    dataset = Dataset()
    dataset.GraphicAnnotationUnits = graphic_annotation_units
    dataset.GraphicDimensions = graphic_dimensions
    dataset.NumberOfGraphicPoints = int(len(graphic_data) / graphic_dimensions)
    dataset.GraphicData = graphic_data
    dataset.GraphicType = graphic_type
    dataset.GraphicFilled = graphic_filled
    dataset.LineStyleSequence = Sequence([line_style])
    if fill_style:
        dataset.FillStyleSequence = Sequence([fill_style])
    return dataset


def get_VOI_LUT_dataset(
        input_dicom,
        referenced_image_sequence,
    ):
    dataset = Dataset()
    dataset.ReferencedImageSequence = referenced_image_sequence

    if input_dicom.__contains__('WindowWidth') and input_dicom.__contains__('WindowCenter'):
        dataset.WindowCenter = input_dicom.WindowCenter
        dataset.WindowWidth = input_dicom.WindowWidth
    elif input_dicom.__contains__('VOILUTSequence'):
        dataset.VOILUTSequence = input_dicom.VOILUTSequence
    else:
        raise KeyError
    
    return dataset
